/* $Id: ClientSocket.C,v 1.8 2003/04/11 23:23:49 sunday Exp $
 * Copyright (c) 2002, 2003 Joe Sunday <sunday@csh.rit.edu>
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Description:
 *   Class to handle input from a connected socket, queue up
 *   incoming traffic until a line is found, and parse that line.
 *   Created by ListenSocket.
 * 
 * Note:
 *   This should really be broken into ClientSocket and
 *   JukeClientSocket, with the generic client stuff
 *   abstracted out for future use.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>

#include "config.h"
#include "Playlist.h"
#include "ClientSocket.h"
#include "ArgVector.h"
#include "Database.h"

// Playlist and database defined in main.C
extern Playlist *playlist;
extern Database *db;

#define STRCMP( a, b ) ( strcasecmp( a, b ) == 0 )

ClientSocket::ClientSocket( int file_descriptor ) :
    fd( file_descriptor ),
    buf( new char[100] ),
    buf_len( 0 ),
    buf_size( 100 ),
    notify( 0 ),
    obuf( NULL ),
    obuf_len( 0 )
{
    bzero( buf, 100 );

    // Set to non-block
    fcntl( fd, F_SETFL, O_NONBLOCK );

    // We don't want to pass this descriptor on when we fork a
    // decoder
    fcntl( fd, F_SETFD, 1 );
    queueOutput( JUKEBANNER );
}

ClientSocket::~ClientSocket( )
{
    shutdown( fd, SHUT_RDWR );
    close( fd );
}

int ClientSocket::getFd() const {
    return fd;
}

int ClientSocket::read() {
    char b[100];
    int l;

    l = ::read( fd, &b, 99 );
   
    while( l > 0 ) {
        if( ( buf_len  + l ) > buf_size ) {
            char *p( new char[buf_size + 100] );
            memcpy( p, buf, buf_size );
            buf_size += 100;
            delete [] buf;
            buf = p;
        }
        memcpy( &buf[buf_len], b, l );
        buf_len += l;
        l = ::read( fd, &b, 99 );
    }
    if( ( l == 0 ) || ( errno != EAGAIN && errno != 0 ) ) {
        return -1;
    }

    buf[ buf_len ] = 0;
    return processBuf();
} 

int ClientSocket::write() {
    int l( 0 );
    if( !obuf ) {
        return 0;
    }

    l = send( fd, obuf, obuf_len, 0 );
    if( l == obuf_len ) {
	obuf_len = 0;
	delete [] obuf;
	obuf = NULL;
    }
    else if( l > 0 ) {
	char *b = new char[obuf_len - l + 1];
	memcpy( b, &obuf[l], obuf_len - l );
	delete [] obuf;
	obuf = b;
	obuf_len -= l;
        obuf[obuf_len] = 0;
    }
    else if( ( l == 0 ) || ( ( l == -1 ) && ( errno != EAGAIN ) ) ) 
        return -1;
    return 0;
}

int ClientSocket::hasOutput() const {
    return obuf_len != 0;
}

void ClientSocket::SetNotify( unsigned int n ) {
    notify |= n;
}

void ClientSocket::cleanBuf() {
    char *a( buf );
    char *n( new char[ buf_size ] );
    char *b( n );
    int newlen( 0 );

    while( *a ) {
        if( *a != '\r' ) {
            *b = *a;
            ++b;
            ++newlen;
        }
        ++a;
    }
    *b = 0;
    delete [] buf;
    buf = n;
    buf_len = newlen;
} 
    
int ClientSocket::processBuf() {
    int l( 0 );
  
    /* Strip out "\r" if it's there. */ 
    cleanBuf();
 
    while( 1 ) {
        l = 0;
        while( buf[l] && buf[l] != '\n' ) l++;
    
        if( buf[l] == '\n' ) {
            char *line = new char[l+1];
            char *b;
            strncpy( line, buf, l );
            line[l] = 0;
            
            if( processLine( line ) ) {
                delete [] line;
                return -1;
            }
            
            if( l < buf_size ) {
                b = new char[buf_size - l];
                memcpy( b, &buf[l+1], buf_size - l );
                buf_size = buf_size - l;
                buf_len = buf_len - l - 1;
                delete [] buf;
                buf = b;
             }
             delete [] line;
        }
        else {
            return 0;
        }
    }
}

int ClientSocket::processLine( char * line ) {
    ArgVector argv( line );
    vector<Song*> *songs = NULL;

    if( !argv.size() )
        return 0;
    if( STRCMP( argv[0], "ADD" ) ) {
        argv.Pop();
        if( argv.size() > 1 && (
            argv.HashFind( "ALBUM" ) || 
            argv.HashFind( "ARTIST" ) ||
            argv.HashFind( "TITLE" ) ||
            argv.HashFind( "ATITLE" ) ) )
        {
            songs = db->SearchTracks( argv );
            if( songs ) {
                playlist->QueueTrack( songs );
                delete songs;
                songs = NULL;
                queueOutput( "+OK\n" );
            } else {
                queueOutput( "-ERR NO TRACKS FOUND\n" );
            }
        } else {
            queueOutput( "-ERR INSUFFICIENT DATA\n" );
        }
    }
    else if( STRCMP( argv[0], "LOAD" ) ) {
        argv.Pop();
        const char *name = argv.HashFind( "NAME" );
        if( name ) {
            playlist->LoadPlaylist( name );
            queueOutput( "+OK\n" );
        }
        else {
            queueOutput( "-ERR INSUFFICIENT INFORMATION\n" );
        }
    }
    else if( STRCMP( argv[0], "SAVE" ) ) {
        argv.Pop();
        const char *name = argv.HashFind( "NAME" );
        if( name ) {
            playlist->SavePlaylist( name );
            queueOutput( "+OK\n" );
        }
        else {
            queueOutput( "-ERR INSUFFICIENT INFORMATION\n" );
        }
    }
    else if( STRCMP( argv[0], "ALBUM" ) ) {
        argv.Pop();
        ListAlbums( argv );
    } 
    else if( STRCMP( argv[0], "TRACK" ) ) {
        argv.Pop();
        ListTracks( argv );
    }
    else if( STRCMP( argv[0], "LIST" ) ) {
        ListTracks( playlist->GetList() );
        notify &= ~NOTIFY_PLAYLIST;
    }
    else if( STRCMP( argv[0], "PLAY" ) ) {
        argv.Pop();
        if( argv.size() ) {
            const char *p = argv.HashFind( "INDEX" );
            if( p ) {
                playlist->ChangeTrack( atoi( p ) );
                queueOutput( "+OK\n" );
            }
            else {
                queueOutput( "-ERR BAD ARGUMENT\n" );
            }
        } else {
            playlist->Resume();
            queueOutput( "+OK\n" );
        }
    }
    else if( STRCMP( argv[0], "SHUFFLE" ) ) {
        playlist->Shuffle();
        queueOutput( "+OK\n" );
    }
    else if( STRCMP( argv[0], "STATUS" ) ) {
        double position;
        int index;
        int status;
        char scr[25];
        status = playlist->Status( &index, &position );
        switch( status ) {
          case DECODE_STOP:
            queueOutput( "STOP " );
            break;
          case DECODE_PAUSE:
            queueOutput( "PAUS " );
            break;
          case DECODE_PLAY:
            queueOutput( "PLAY " );
            break;
        }
        sprintf( scr, "%d %.3f", index, position );
        queueOutput( scr );
        if( (notify & NOTIFY_PLAYLIST) == NOTIFY_PLAYLIST ) {
            queueOutput( " PLAYLIST" );
        }
        queueOutput( "\n+OK\n" );
    }
    else if( STRCMP( argv[0], "PAUSE" ) ) {
        playlist->Pause();
        queueOutput( "+OK\n" );
    }
    else if( STRCMP( argv[0], "MOVE" ) ) {
        argv.Pop();
        const char *o = argv.HashFind( "OLD" );
        const char *n = argv.HashFind( "NEW" );
        if( !o || !n ) {
            queueOutput( "-ERR BAD ARGUMENT\n" );
        } else {
            playlist->Move( atoi( o ), atoi( n ) );
            queueOutput( "+OK\n" );
        }
    }
    else if( STRCMP( argv[0], "REMOVE" ) ) {
        argv.Pop();
        const char *index = argv.HashFind( "INDEX" );
        if( !index ) {
            queueOutput( "-ERR BAD ARGUMENT\n" );
        } else {
            playlist->Remove( atoi( index ) );
            queueOutput( "+OK\n" );
        }
    } 
    else if( STRCMP( argv[0], "CLEAR" ) ) {
        playlist->Clear();
        queueOutput( "+OK\n" );
    } 
    else if( STRCMP( argv[0], "SETUP" ) ) {
        argv.Pop();
        switch( playlist->GetEndMode() ) {
          case Playlist::END_STOP:
            queueOutput( "END STOP " );
            break;
          case Playlist::END_LOOP:
            queueOutput( "END LOOP " );
            break;
          case Playlist::END_RANDOM:
            queueOutput( "END RANDOM " );
            break;
            
        }
        switch( playlist->GetFifoMode() ) {
          case Playlist::FIFO_ON:
            queueOutput( "FIFO ON " );
            break;
          case Playlist::FIFO_OFF:
            queueOutput( "FIFO OFF" );
            break;
        }
        queueOutput( "\n+OK\n" );
    }
    else {
        queueOutput( "-ERR Unknown Command " );
        queueOutput( argv[0] );
        queueOutput( "\n" );
    }
    return 0;
}

void ClientSocket::queueOutput( const char *output ) {
    char *b( new char[ strlen( output ) + obuf_len + 1] );

    *b = 0;
    if( obuf ) {
        memcpy( b, obuf, obuf_len + 1 );
        delete [] obuf;
        obuf = NULL;
    }
    obuf = b;
    strcat( obuf, output );
    obuf_len += strlen( output );
}

void ClientSocket::ListTracks( const vector<Song*> *list ) {
    unsigned int i;
    char *line;
    if( list ) {
        for( i = 0; i < list->size(); i++ ) {
            char *album_title = AddSlashes( (*list)[i]->GetAlbumTitle() );
            char *title = AddSlashes( (*list)[i]->GetTitle() );
            char *artist = AddSlashes( (*list)[i]->GetArtist() );
            line = (char*)malloc( 80 + strlen( album_title ) + strlen( title ) + strlen( artist ) );
            sprintf( line, "%d %d \"%s\" \"%s\" \"%s\" %d\n",
                (*list)[i]->GetAlbum(),
                (*list)[i]->GetTrack(),
                album_title,
                title,
                artist,
                (int)(*list)[i]->GetLength() );
            queueOutput( line );
            free( line );
            free( album_title );
            free( title );
            free( artist );
        }
    }
    queueOutput( "+OK\n" );
}

void ClientSocket::ListTracks( const ArgVector &argv ) {
    vector<Song*> *list = db->SearchTracks( argv );
    ListTracks( list );
    if( list ) {
        while( !list->empty() ) {
            delete (*list)[0];
            list->erase( list->begin() );
        }
        delete list;
    }
}

void ClientSocket::ListAlbums( const vector<Album*> *list ) {
    unsigned int i;
    char *line;
    if( list ) {
        for( i = 0; i < list->size(); i++ ) {
            char *title = AddSlashes( (*list)[i]->GetTitle() );
            char *artist = AddSlashes( (*list)[i]->GetArtist() );
            line = (char*)malloc( 80 + strlen( title ) + strlen( artist ) );
            sprintf( line, "%d \"%s\" \"%s\" %d %d\n",
                (*list)[i]->GetAlbum(),
                title,
                artist,
                (int)(*list)[i]->GetLength(),
                (*list)[i]->GetTracks() );
            queueOutput( line );
            free( line );
            free( title );
            free( artist );
        }
    }
    queueOutput( "+OK\n" );
}

void ClientSocket::ListAlbums( const ArgVector &argv ) {
    vector<Album*> *list = db->SearchAlbums( argv );
    ListAlbums( list );
    if( list ) {
        while( !list->empty() ) {
            delete (*list)[0];
            list->erase( list->begin() );
        }
        delete list;
    }
}

char *ClientSocket::AddSlashes( const char *string ) {
    char *newstring = (char*)malloc( (strlen( string ) * 2 ) + 1);
    const char *p = string;
    char *q = newstring;
   
    while( *p ) {
        if( *p == '"' || *p == '\\' ) {
            *q++ = '\\';
        }
        *q++ = *p++;
    }
    *q = 0;
    return newstring;
}
