/* $Id: Database.C,v 1.7 2003/04/13 08:58:37 sunday Exp $
 * Copyright (c) 2002, 2003 Joe Sunday <sunday@csh.rit.edu>
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Description:
 *   Class to handle database connection and queries.
 *
 */

#include <iostream>

#include "config.h"
#include "Database.h"

#define WHERE( a ) ((((*a)++)==0)?"WHERE":"AND")

Database::Database() :
    conn( PQconnectdb( DBSTRING ) )
{
    if( PQstatus( conn ) != CONNECTION_OK ) {
        cerr << "Error on DB Connect: " << PQerrorMessage( conn ) << endl;
        exit( 1 );
    }
}

Database::~Database() {
    PQfinish( conn ); 
}

Song *Database::CreateSong( unsigned int album, unsigned int track )
{
    PGresult *res;
    char query[256];

    sprintf( query, 
      "SELECT album_title, title, artist, date_part( 'epoch', length ) " \
      "FROM track_detail WHERE album = %d AND track = %d", 
      album, track );

    res = PQexec( conn, query );
    if( res && 
        PQresultStatus( res ) == PGRES_TUPLES_OK && 
        PQntuples( res ) == 1 ) 
    {
        Song *p = new Song( album, track, 
          PQgetvalue( res, 0, 0 ),
          PQgetvalue( res, 0, 1 ),
          PQgetvalue( res, 0, 2 ),
          atoi( PQgetvalue( res, 0, 3 ) ) );
        PQclear( res );
        return p;
    }
    else if( res ) {
        PQclear( res );
    }
    return NULL;
}

vector<Song*> *Database::SearchTracks( const ArgVector &argv )
{
    const char *album = argv.HashFind( "ALBUM" );
    const char *track = argv.HashFind( "TRACK" );
    const char *artist = argv.HashFind( "ARTIST" );
    const char *title = argv.HashFind( "TITLE" );
    const char *atitle = argv.HashFind( "ATITLE" );

    vector<Song*> *v = NULL;

    int and_clause = 0;
    char *where = NULL;
    int x = 0;
    if( album ) {
        x = WhereInt( &where, "album", atoi( album ), &and_clause );
    }

    if( track ) {
        x = WhereInt( &where, "track", atoi( track ), &and_clause );
    }

    if( artist ) {
        x = WhereString( &where, "artist", artist, &and_clause );
    }

    if( title ) {
        x = WhereString( &where, "title", title, &and_clause );
    }

    if( atitle ) {
        x = WhereString( &where, "album_title", atitle, &and_clause );
    }

    v = TrackQuery( where );

    free( where );
    return v;
}

Song *Database::RandomTrack() {
    Song *s = NULL;
    vector<Song*> *v = TrackQuery( "ORDER BY RANDOM() LIMIT 1" );
    if( v && v->size() ) {
        s = (*v)[0];
    }
    if( v ) {
        delete v;
    }
    return s;
}

vector<Album*> *Database::SearchAlbums( const ArgVector &argv ) 
{
    char *query = NULL;
    const char *id = argv.HashFind( "ALBUM" );
    const char *artist = argv.HashFind( "ARTIST" );
    const char *title = argv.HashFind( "TITLE" );

    char *where = NULL;
    int and_clause = 0;
    int x = 0;
    vector<Album*> *v = NULL;

    PGresult *res;

    if( id ) {
        x = WhereInt( &where, "album", atoi( id ), &and_clause );
    }

    if( artist ) {
        x = WhereString( &where, "artist", artist, &and_clause );
    }

    if( title ) {
        x = WhereString( &where, "title", title, &and_clause );
    }

    query = (char*)malloc( 160 + x );
    x = sprintf( query, "SELECT album, artist, title, " \
        "date_part( 'epoch', length), tracks FROM album_detail " );
    if( where ) {
        strlcat( query, where, 160+x );
    }
    res = PQexec( conn, query );
    if( res && 
        PQresultStatus( res ) == PGRES_TUPLES_OK && 
        PQntuples( res ) > 0 ) 
    {
        v = new vector<Album*>;
        for( x = 0; x < PQntuples( res ); x++ ) {
            Album *a = new Album( 
              atoi( PQgetvalue( res, x, 0 ) ),
              PQgetvalue( res, x, 1 ),
              PQgetvalue( res, x, 2 ),
              atoi( PQgetvalue( res, x, 3 ) ),
              atoi( PQgetvalue( res, x, 4 ) ) );
            v->push_back( a );
        }
    }
    if( res ) {
        PQclear( res );
    }
    if( where ) {
        free( where );
    }
    free( query );
    return v;
}

vector<Song*> *Database::LoadPlaylist( const char *name, unsigned int *index )
{
    char *query = NULL;
    int x = strlen( name ) * 2 + 1;
    char *name2 = (char*)malloc( x );
    x = PQescapeString( name2, name, x );
    vector<Song*> *v = NULL;
    PGresult *res;

    query = (char*)malloc( 360 + x );
    sprintf( query, "SELECT last FROM playlists WHERE name = '%s'", name2 );

    res = PQexec( conn, query );
    if( res && 
        PQresultStatus( res ) == PGRES_TUPLES_OK && 
        PQntuples( res ) > 0 ) 
    {
        *index = atoi( PQgetvalue( res, 0, 0 ) );
        PQclear( res );
    }
    else {
        if( res ) {
            PQclear( res );
        }
        free( query );
        free( name2 );
        cerr << "Could't load index." << endl;
        return NULL;
    } 
    
    sprintf( query, 
      "SELECT a.album, a.track, a.album_title, a.title, a.artist, " \
      "date_part( 'epoch', a.length ) FROM track_detail a, playlist_entries p " \
      "WHERE p.name = '%s' AND a.album = p.album AND a.track = p.track " \
      "ORDER BY p.position", name2 );
 
    res = PQexec( conn, query );
    if( res && 
        PQresultStatus( res ) == PGRES_TUPLES_OK && 
        PQntuples( res ) > 0 ) 
    {
        v = new vector<Song*>;
        for( x = 0; x < PQntuples( res ); x++ ) {
            Song *p = new Song(
              atoi( PQgetvalue( res, x, 0 ) ),
              atoi( PQgetvalue( res, x, 1 ) ),
              PQgetvalue( res, x, 2 ),
              PQgetvalue( res, x, 3 ),
              PQgetvalue( res, x, 4 ),
              atoi( PQgetvalue( res, x, 5 ) ) );
            v->push_back( p );
        }
        PQclear( res );
    }
    else if( res && PQresultStatus( res ) != PGRES_TUPLES_OK ) {
        cerr << "Could't load entries: " << PQresultErrorMessage( res ) << endl;
        PQclear( res );
    }
    else if( res ) {
        PQclear( res );
    }
    free( name2 );
    free( query );
    return v;
}

int Database::SavePlaylist( const char *name, vector<Song*> *list, unsigned int index )
{
    char *query = NULL;
    int x = strlen( name ) * 2 + 1;
    unsigned int i;
    char *name2 = (char*)malloc( x );
    PGresult *res;

    x = PQescapeString( name2, name, x );

    res = PQexec( conn, "BEGIN" );
    if( !res || PQresultStatus( res ) != PGRES_COMMAND_OK ) {
        cerr << "Couldn't start transaction: ";
        if( res ) {
            cerr << PQresultErrorMessage( res ) << endl;
            PQclear( res );
        }
        else {
            cerr << "Unknown error." << endl;
        }
        return -1;
    }
    PQclear( res );

    query = (char*)malloc( 1024 + x );
    sprintf( query, "DELETE FROM playlists WHERE name = '%s'", name2 );
 
    res = PQexec( conn, query );
    if( !res || PQresultStatus( res ) != PGRES_COMMAND_OK ) {
        free( query );
        cerr << "Delete playlist failed: ";
        if( res ) {
            cerr << PQresultErrorMessage( res ) << endl;
            PQclear( res );
        }
        else {
            cerr << "Unknown error." << endl;
        }
        PQexec( conn, "ROLLBACK" );
        return -1;
    }
    PQclear( res );

    sprintf( query, "INSERT INTO playlists( name, last ) " \
      "VALUES( '%s', %d )", name2, index );

    res = PQexec( conn, query );
    if( !res || PQresultStatus( res ) != PGRES_COMMAND_OK ) {
        free( query );
        cerr << "Create playlist failed: ";
        if( res ) {
            cerr << PQresultErrorMessage( res ) << endl;
            PQclear( res );
        }
        else {
            cerr << "Unknown error." << endl;
        }
        PQexec( conn, "ROLLBACK" );
        return -1;
    }
    PQclear( res );

    for( i = 0; i < list->size(); i++ ) {
        sprintf( query, 
          "INSERT INTO playlist_entries( name, position, album, track ) " \
          "VALUES( '%s', %d, %d, %d)", name2, i, 
          (*list)[i]->GetAlbum(), (*list)[i]->GetTrack() );
        res = PQexec( conn, query );
        if( !res || PQresultStatus( res ) != PGRES_COMMAND_OK ) {
            free( query );
            cerr << "Insert failed.\n";
            PQexec( conn, "ROLLBACK" );
            return -1;
        }
        PQclear( res );
    }
    free( query );
    res = PQexec( conn, "COMMIT" );
    PQclear( res );
    return 0;
}

vector<Song*> *Database::TrackQuery( const char *where ) {
    vector<Song*> *v = NULL;
    char *query;
    PGresult *res;

    query = (char*)malloc( 180 + (where?strlen( where ):0) );

    sprintf( query, 
      "SELECT album, track, album_title, title, artist, " \
      "date_part( 'epoch', length ) FROM track_detail " );

    if( where ) {
        strcat( query, where );
    }
 
    res = PQexec( conn, query );
    if( res && 
        PQresultStatus( res ) == PGRES_TUPLES_OK && 
        PQntuples( res ) > 0 ) 
    {
        v = new vector<Song*>;
        for( int x = 0; x < PQntuples( res ); x++ ) {
            Song *p = new Song(
              atoi( PQgetvalue( res, x, 0 ) ),
              atoi( PQgetvalue( res, x, 1 ) ),
              PQgetvalue( res, x, 2 ),
              PQgetvalue( res, x, 3 ),
              PQgetvalue( res, x, 4 ),
              atoi( PQgetvalue( res, x, 5 ) ) );
            v->push_back( p );
        }
        PQclear( res );
    }
    else if( res ) {
        PQclear( res );
    }
    free( query );
    return v;
}

int Database::WhereString( char **where, const char *column,
  const char *value, int *prefix ) {
    int x;
    int l = (*where)?strlen( *where ):0;
    char *scratch, *scratch2;

    x = strlen( value ) * 2 + 1;
    scratch = (char*)malloc( x );
    x = PQescapeString( scratch, value, x );

    scratch2 = (char*)malloc( 80 + strlen( column ) + x );
    x = sprintf( scratch2, "%s %s ILIKE '%s' ", WHERE(prefix), 
      column, scratch );
    l += x + 1;
    if( *where ) {
        *where = (char*)realloc( *where, l );
        strlcat( *where, scratch2, l );
        free( scratch2 );
    } else {
        *where = scratch2;
    }
    free( scratch );
    return l;
}

int Database::WhereInt( char **where, const char *column,
  int value, int *prefix ) {
    int x = 0;
    int l = (*where)?strlen( *where ):0;
    char *scratch = (char*)malloc( 80 + strlen( column ) );

    x = sprintf( scratch, "%s %s=%d ", WHERE(prefix), column, value );
    l += x + 1;
    if( *where ) {
        *where = (char*)realloc( *where, l );
        strlcat( *where, scratch, l );
        free( scratch );
    } else {
        *where = scratch;
    }
    return l;
}

