/* $Id: Decode.C,v 1.4 2003/04/11 23:32:53 sunday Exp $
 * Copyright (c) 2002, 2003 Joe Sunday <sunday@csh.rit.edu>
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Description:
 *   Class to handle asynchronous OggVorbis decoding.
 *
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
#include <string.h>
#include <iostream>
#include <pthread.h>

#include "config.h"
#include "Decode.h"
#include "sound.h"

#define BUF_SIZE 1024

char *OggError::text[] = { "Couldn't Open File", "Ogg Vorbis Error" };
OggError::OggError( OggError::error e ) :
    err( e )
{
}

const char *OggError::ToString() const {
    return text[(int)err];
}

OggFile::OggFile( int a, int t ) throw( OggError ) :
    album( a ),
    track( t ),
    section( -1 ),
    position( 0 ),
    endian( sound_vorbis_endian() )
{
    char fname[sizeof( MUSICROOT ) + 20];
    FILE *file;
 
    sprintf( fname, "%s/%.8x/%.2x.ogg", MUSICROOT, album, track );
    file = fopen( fname, "r" );
    if( !file ) {
        throw( OggError( OggError::NO_FILE ) );
    }

    if( ov_open( file, &vf, NULL, 0 ) < 0 ) {
        fclose( file );
        throw( OggError( OggError::NO_OGG ) );
    }

    length = ov_time_total( &vf, -1 );
}

OggFile::~OggFile() {
    ov_clear( &vf );
}

int OggFile::Read( char *buf, unsigned int len ) {
    int ret = 0;
    int l;
 
    do {
        l = ov_read( &vf, &buf[ret], len - ret, endian, 2, 1, &section );
        ret += l;
    } while( (l > 0) && (ret < len) );
        
    position = ov_time_tell( &vf );

    return ret;
}

int OggFile::GetAlbum() const {
    return album;
}

int OggFile::GetTrack() const {
    return track;
}

double OggFile::GetPosition() const {
    return position;
}

double OggFile::GetLength() const {
    return length;
}

int OggFile::GetSection() const {
    return section;
}

int OggFile::GetChannels() {
    vorbis_info *vi = ov_info( &vf, -1 );
    return vi->channels;
}

int OggFile::GetRate() {
    vorbis_info *vi = ov_info( &vf, -1 );
    return vi->rate;
}

/* Decoder Class
 *
 *
 */

static void *decode_stub( void *arg ) {
    return ((OggDecoder*)arg)->Thread();
}

OggDecoder::OggDecoder() :
    current( NULL ),
    next( NULL ),
    next_callback( NULL ),
    next_callback_arg( NULL ),
    action( ACTION_STOP )
{
    pthread_attr_t attr;
    pthread_mutexattr_t mattr;
    pthread_mutexattr_init( &mattr );
    pthread_mutexattr_settype( &mattr, PTHREAD_MUTEX_RECURSIVE );

    sound_init();

    /* Setup mutexes and cond */
    pthread_mutex_init( &next_mutex, &mattr );
    pthread_mutex_init( &action_mutex, NULL );
    pthread_cond_init( &action_cond, NULL );

    pthread_attr_init( &attr );
    pthread_attr_setschedpolicy( &attr, SCHED_RR );
    pthread_create( &decode_thread, &attr, decode_stub, (void*)(this) );
}

OggDecoder::~OggDecoder() {
}

void *OggDecoder::Thread() {
    char buf[BUF_SIZE];
    int section = -1;
    int ret;

    while( 1 ) {
        pthread_mutex_lock( &action_mutex );

        switch( action ) {
          case ACTION_NEXT:
            pthread_mutex_lock( &next_mutex );
            if( current ) {
               delete current;
            }
            current = next;
            next = NULL;
            section = -1;
            pthread_mutex_unlock( &next_mutex );
            action = ACTION_CONTINUE;

          case ACTION_CONTINUE:
            if( !current ) {
                action = ACTION_STOP;
                pthread_mutex_unlock( &action_mutex );
                if( next_callback != NULL ) {
                    next_callback( next_callback_arg );
                }
                pthread_mutex_lock( &action_mutex );
                pthread_cond_wait( &action_cond, &action_mutex );
                break;
            }
            pthread_mutex_unlock( &action_mutex );
            ret = current->Read( buf, BUF_SIZE );
            if( ret == 0 ) {
                pthread_mutex_lock( &action_mutex );
                delete current;
                current = NULL;
                action = ACTION_STOP;
                pthread_mutex_unlock( &action_mutex );
                if( next_callback != NULL ) {
                    next_callback( next_callback_arg );
                }
                pthread_mutex_lock( &action_mutex );
                if( action == ACTION_STOP ) {
                    pthread_cond_wait( &action_cond, &action_mutex );
                }
                break;
            } else {
                /* Check if section numbers changed */
                if( section != current->GetSection() ) {
                    sound_setup( current->GetChannels(), current->GetRate() );
                    section = current->GetSection();
                };
                sound_play( buf, ret );
                pthread_mutex_lock( &action_mutex );
            }
            break;
     
          case ACTION_STOP:
            pthread_cond_wait( &action_cond, &action_mutex );
            break;
        }
        pthread_mutex_unlock( &action_mutex );
    }
    return NULL;
}

int OggDecoder::RegisterNextCallback( void(*cb)(void *), void *arg ) {
    next_callback = cb;
    next_callback_arg = arg;
    return 0;
}

int OggDecoder::ChangeTrack( int album, int track ) {
    /* Grab the next mutex */
    try {
        OggFile *f = new OggFile( album, track );
        pthread_mutex_lock( &next_mutex );
        if( next ) {
            delete next;
        }
        next = f;
        pthread_mutex_unlock( &next_mutex );
        pthread_mutex_lock( &action_mutex );
        action = ACTION_NEXT;
        pthread_mutex_unlock( &action_mutex );
 
        /* Signal the decoder thread to wake up if it's sleeping */
        pthread_cond_signal( &action_cond );
        return 0;
    } catch( OggError e ) {
        return -1;
    }
}

int OggDecoder::Pause() {
    pthread_mutex_lock( &action_mutex );
    action = ACTION_STOP;
    pthread_mutex_unlock( &action_mutex );
    return 0;
}

int OggDecoder::Resume() {
    pthread_mutex_lock( &action_mutex );
    action = ACTION_CONTINUE;
    pthread_mutex_unlock( &action_mutex );
    pthread_cond_signal( &action_cond );
    return 0;
}

int OggDecoder::Status( int *album, int *track, double *pos, double *len ) {
    pthread_mutex_lock( &action_mutex );
    if( !current ) {
        pthread_mutex_unlock( &action_mutex );
        return DECODE_STOP;
    }

    *album = current->GetAlbum();
    *track = current->GetTrack();
    *pos = current->GetPosition();
    *len = current->GetLength();
    pthread_mutex_unlock( &action_mutex );
    if( action == ACTION_CONTINUE ) {
        return DECODE_PLAY;
    }
    return DECODE_PAUSE;
}
