/* $Id: Playlist.C,v 1.9 2003/04/11 23:32:53 sunday Exp $
 * Copyright (c) 2002, 2003 Joe Sunday <sunday@csh.rit.edu>
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Description:
 *   Class to manage playlist and feed current track to decoder.
 *
 */

#include <iostream>

#include <stdlib.h>
#include <string.h>
#include <algorithm>

#include "Playlist.h"
#include "Database.h"

extern Database *db;

static void next_callback( void *arg ) {
    ((Playlist*)arg)->Next();
}

Playlist::Playlist( const char *p, end_modes e, fifo_modes f ) :
    persist( NULL ),
    current_track( 0 ),
    notify( 0 ),
    end_mode( e ),
    fifo_mode( f )
{
    pthread_mutexattr_t a;
    pthread_mutexattr_init( &a );
    pthread_mutexattr_settype( &a, PTHREAD_MUTEX_RECURSIVE );
    d.RegisterNextCallback( next_callback, (void*)this );
    pthread_mutex_init( &list_mutex, &a );
    if( p ) {
        persist = strdup( p );
        LoadPlaylist( persist );
    }
}

Playlist::~Playlist() {
    unsigned int i;

    d.RegisterNextCallback( NULL, NULL );
    if( persist ) {
        SavePlaylist( persist );
        free( persist );
    }
    pthread_mutex_lock( &list_mutex );
    for( i = 0; i < list.size(); i++ ) {
        delete list[i];
    }
    pthread_mutex_destroy( &list_mutex );
}

void Playlist::LoadPlaylist( const char *name ) {
    unsigned int index;
    vector<Song*> *s = db->LoadPlaylist( name, &index );
    if( s ) {
        Clear();
        QueueTrack( s );
        delete s;
        ChangeTrack( index );
    }
}


void Playlist::SavePlaylist( const char *name ) {
    pthread_mutex_lock( &list_mutex );
    db->SavePlaylist( name, &list, current_track );
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::Shuffle( ) {
    Song *p = NULL;

    pthread_mutex_lock( &list_mutex );
    if( list.size() < 2 ) {
        pthread_mutex_unlock( &list_mutex );
        return;
    }

    // Save the current song
    if( current_track < list.size() ) {
        p = list[current_track];
        list.erase( list.begin() + current_track );
    }

    random_shuffle( list.begin(), list.end() );
    /* 
     * Rather than incur the loss of inserting in front, we move
     * the current last track to the end.
     */
    if( p ) {     
        Song *p2 = list[0];
        list[0] = p;
        list.push_back( p2 );
        current_track = 0;
    }

    notify |= NOTIFY_PLAYLIST;
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::QueueTrack( Song *track ) {
    if( !track ) {
        return;
    }

    pthread_mutex_lock( &list_mutex );
    list.push_back( track );

    if( current_track >= list.size() - 1 ) { 
        ChangeTrack( list.size() - 1 );
    }
    notify |= NOTIFY_PLAYLIST;
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::QueueTrack( const vector<Song*> *tracks ) {
    if( !tracks ) {
        return;
    }

    for( unsigned int i = 0; i < tracks->size(); i++ ) {
        QueueTrack( (*tracks)[i] );
    }
}
    
void Playlist::Remove( unsigned int index ) {
    pthread_mutex_lock( &list_mutex );
    if( index >= list.size() ) {
        pthread_mutex_unlock( &list_mutex );
        return;
    }

    Song *p = list[index];
    list.erase( list.begin() + index );
    delete p;
    notify |= NOTIFY_PLAYLIST;

    if( index == current_track ) {
        ChangeTrack( current_track );
    } 
    else if( current_track > index ) {
        current_track -= 1;
    }
     
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::Clear() {
    pthread_mutex_lock( &list_mutex );
    d.Pause();
    while( !list.empty() ) {
        Song *p = list[0];
        list.erase( list.begin() );
        delete p;
    }
    current_track = 0;
    notify |= NOTIFY_PLAYLIST;
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::Move( unsigned int index, unsigned int new_index ) {
    unsigned int i;
    pthread_mutex_lock( &list_mutex );
    if( index >= list.size() || new_index >= list.size() ) {
        pthread_mutex_unlock( &list_mutex );
        return;
    }

    Song *p = list[index];
    if( index < new_index ) {
        for( i = index; i < new_index; i++ ) {
            list[i] = list[i+1];
        }
        if( current_track > index && current_track <= new_index ) {
            current_track -= 1;
        } 
        else if( current_track == index ) {
            current_track = new_index;
        }
    } else {
        for( i = index; i > new_index; i-- ) {
            list[i] = list[i-1];
        }
        if( current_track < index && current_track >= new_index ) {
            current_track += 1;
        }
        else if( current_track == index ) {
            current_track = new_index;
        }
    }
    list[new_index] = p;

    if( fifo_mode == FIFO_ON && index == 0 || new_index == 0 ) {
        ChangeTrack( 0 );
    }

    notify |= NOTIFY_PLAYLIST;
    pthread_mutex_unlock( &list_mutex );
}

void Playlist::Next() {
    ChangeTrack( current_track + 1 );
}

void Playlist::ChangeTrack( unsigned int index ) {
    unsigned int last_track;
    pthread_mutex_lock( &list_mutex );

    last_track = current_track;
    current_track = index;

    if( ( fifo_mode == FIFO_ON ) && ( current_track != 0 ) && 
        ( current_track != last_track ) ) {
        if( last_track != 0 ) {
            cerr << "index " << last_track << " instead of 0.\n";
        } else if( list.size() == 1 ) {
            Song *s = list[0];
            delete s;
            list.erase( list.begin() );
            notify |= NOTIFY_PLAYLIST;
        } else if( current_track < list.size() ) {
            Song *s = list[0];
            delete s;
            
            list[0] = list[current_track];
            list.erase( list.begin() + current_track );
            current_track = 0;
            notify |= NOTIFY_PLAYLIST;
        }
    }

    if( current_track >= list.size() ) {
        if( ( end_mode == END_LOOP ) && list.size() ) {
            current_track = 0;
        } else if( end_mode == END_RANDOM ) {
            Song *s = db->RandomTrack();
            if( s ) {
                list.push_back( s );
                current_track = list.size() - 1;
                notify |= NOTIFY_PLAYLIST;
            }
        } 
    }

    if( current_track < list.size() ) {
        Song *p = list[current_track];
        d.ChangeTrack( p->GetAlbum(), p->GetTrack() );
    } else {
        d.Pause();
    }
    pthread_mutex_unlock( &list_mutex );
}

int Playlist::Status( int *index, double *position ) {
    int ret = DECODE_STOP;
    double p, l;
    int album, track;

    pthread_mutex_lock( &list_mutex );
    *index = 0;
    *position = 0.00;
    if( current_track < list.size() ) {
        ret = d.Status( &album, &track, &p, &l );
        
        if( ret == DECODE_STOP ) {
            *position = 0;
        } else {
            *index = current_track;
            *position = p;
        }
    }
    pthread_mutex_unlock( &list_mutex );
    return ret;
}

void Playlist::Pause() {
    d.Pause();
}

void Playlist::Resume() {
    d.Resume();
}

enum Playlist::end_modes Playlist::GetEndMode() const {
    return end_mode;
}

enum Playlist::fifo_modes Playlist::GetFifoMode() const {
    return fifo_mode;
}

unsigned int Playlist::GetNotify() {
    unsigned int n;
    pthread_mutex_lock( &list_mutex ); 
    n = notify;
    notify = 0;
    pthread_mutex_unlock( &list_mutex );
    return n;
}

const vector<Song*> *Playlist::GetList() const {
    return &list;
}

