#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#include <iostream>
#include <vector>

#include "config.h"
#include "Playlist.h"
#include "ClientSocket.h"
#include "ListenSocket6.h"
#include "Database.h"

#include "sound.h"

using namespace std;

Database *db( NULL );
Playlist *playlist( NULL );

struct juke_config_t {
    Playlist::end_modes end_mode;
    Playlist::fifo_modes fifo_mode;
    const char *persist;
};

void *run_input( void* arg );

/* 
 * Catch certain signals and call exit(0). The default handlers
 * don't attempt to do any sort of cleanup. Calling exit() makes
 * sure the destructor for playlist gets called and the playlist
 * gets dumped back to the database.
 */
extern "C" {
void kill_handle( int sig );
}

void kill_handle( int sig ) {
    if( playlist ) {
        delete playlist;
    }

    if( db ) {
        delete db;
    }
    exit( 0 );
}

void usage( const char *exe ) {
    cout << "\nUsage:\n";
    cout << exe << " [-fglrs]\n\n";
    cout << "Flags:\n";
    cout << "-f\tFifo Mode (Default)\n";
    cout << "-g\tNon-Fifo Mode\n";
    cout << "-l\tLoop At End\n";
    cout << "-r\tRandom At End (Default)\n";
    cout << "-s\tStop At End\n";
    cout << "Note: -f and -l are mutually exclusive.\n";
}

int main( int argc, char **argv ) {
    static struct juke_config_t juke_config = 
        { Playlist::END_RANDOM, Playlist::FIFO_ON, "last" };

    int ch;

    while( (ch = getopt( argc, argv, "fglrs" )) != -1 ) {
        switch( ch ) {
          case 'l':
            juke_config.end_mode = Playlist::END_LOOP;
            break;
          case 's':
            juke_config.end_mode = Playlist::END_STOP;
            break;
          case 'r':
            juke_config.end_mode = Playlist::END_RANDOM;
            break;
          case 'f':
            juke_config.fifo_mode = Playlist::FIFO_ON;
            break;
          case 'g':
            juke_config.fifo_mode = Playlist::FIFO_OFF;
            break;
          case '?':
          default:
            usage( argv[0] );
            exit( 1 );
        }
    }

    if( ( juke_config.fifo_mode == Playlist::FIFO_ON ) && 
        ( juke_config.end_mode == Playlist::END_LOOP ) ) {
        cerr << "Fifo mode and Loop At End are mutually exclusive.\n";
        exit( 1 );
    }


    signal( SIGINT, kill_handle );
    signal( SIGHUP, kill_handle );

    run_input( (void*)&juke_config );
    return 0;
}

void *run_input( void *arg ) {
    fd_set r, w;
    unsigned int notify;
    vector<ClientSocket *>clients;
    struct juke_config_t *juke_config = (struct juke_config_t*)arg;

    struct timeval tv;
    try {
        ListenSocket6 ls( JUKEPORT );
        db = new Database();
        playlist = new Playlist( juke_config->persist, 
            juke_config->end_mode, juke_config->fifo_mode );

        while( 1 ) {
            tv.tv_sec = 5;
            tv.tv_usec = 0;
         
            vector<ClientSocket *>::iterator it;
            int maxfd = ls.getFd();

            FD_ZERO( &r );
            FD_ZERO( &w );

            FD_SET( ls.getFd(), &r );

            if( !clients.empty() ) {
                for( it = clients.begin(); it != clients.end(); ++it ) {
                    FD_SET( (*it)->getFd(), &r );
                    FD_SET( (*it)->getFd(), &w );
                    maxfd = ((*it)->getFd() > maxfd)?(*it)->getFd():maxfd;
                 }
            }
            
            select( maxfd + 1, &r, &w, NULL, NULL );

            if( FD_ISSET( ls.getFd(), &r ) ) {
                ClientSocket *s = ls.accept();
                clients.push_back( s );
            }

            if( !clients.empty() ) {
                for( it = clients.begin(); (!clients.empty()) && 
                     (it != clients.end()); ) {
                    vector< ClientSocket* >::iterator next = it + 1;
                    if( FD_ISSET( (*it)->getFd(), &r ) ) {
                        if( (*it)->read() ) {
                            delete *it;
                            clients.erase( it );
                            continue;
                        }
                    }
                    it = next;
                }
            }
            if( !clients.empty() ) {
                for( it = clients.begin(); (!clients.empty()) && 
                     (it != clients.end()); ) {
                    vector< ClientSocket* >::iterator next = it + 1;
                    if( FD_ISSET( (*it)->getFd(), &w )  && (*it)->hasOutput() )
{
                        if( (*it)->write() ) {
                            delete *it;
                            clients.erase( it );
                            continue;
                        }
                    }
                    it = next;
                }
            }

            // Set notify flags
            if( (notify = playlist->GetNotify()) ) {
                for( it = clients.begin(); (!clients.empty()) &&
                     (it != clients.end()); it = it + 1 ) {
                    (*it)->SetNotify( notify );
                }
            } 
        }
    }
    catch( ListenSocket::Error e ) {
        cerr << "Bailing\n";
        exit( 1 );
    }
    return NULL;
}
